////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// File:     hrcookie.js
// Author:   Henk Reints - Copyright (c) 2007
// Web site: http://www.henk-reints.nl
// Date:     March 2007
// Purpose:  Define JavaScript Objects to hold info of MS Internet Explorer Cookies
//
// DISCLAIMER:
//	HENK REINTS HAS WRITTEN THIS SOFWARE FOR HIS OWN USE, AS A HOBBY, WITHOUT ANY COMMERCIAL INTENTIONS.
//	IT IS PROVIDED AS IS, WITHOUT ANY GUARANTEE FOR CORRECTNESS OR USABILITY. USAGE IS AT YOUR OWN RISK.
//	IN NO WAY SHALL HENK REINTS BE LIABLE FOR ANY NEGATIVE CONSEQUENCES OF USING THIS SOFTWARE.
//
// Objects:
//	HRCookieList	searches all cookies, optionally filtered by a regular expression to match the cookie name
//			and contains the list of all cookies found
//	HRCookie	contains an individual cookie, i.e. the interpreted content of an individual cookie file
//	HRCookieItem	contains interpreted individual name=value items as contained in the cookie file
//
// Usage examples:
//	var myCookieList = new HRCookieList()	// find ALL cookies
//	WScript.StdOut.WriteLine (myCookieList)	// list ALL cookie info on StdOut
//
//	var filteredCookieList = new HRCookieList(/[@|\.]henk-reints\.nl\//i)	// find all cookies of henk-reints.nl
//	WScript.Echo (filteredCookieList)					// and show them
//
// About index.dat in the Cookies folder:
//	it contains a list of cookie names and corresponding cookie file names, plus some extra info
//	Found info on http://codespeak.net/svn/wwwsearch/ClientCookie/trunk/ClientCookie/_MSIECookieJar.py
//	and via http://www.wotsit.org and by watching what happens in the file when visiting various sites.
//	General structure as far I know now:
//	- first a 32 byte signature matching the regular expression: /Client UrlCache MMF Ver \d\.\d.*/
//	- then a dword holding the offset (from start of file) of the first record
//	    (or, what's the same, the size of this header record)
//	each record is a number of 128 byte blocks and contains:
//	- a 4 byte signature ("HASH", "URL ", or 4 nullbytes marking logical end-of-file)
//	- a dword holding the length of the record in 128 byte blocks
//	- other data
//	I do not know what the HASH records are
//	URL records seem to contain:
//	- the "URL " signature
//	- the dword record length in 128 byte blocks
//	- a quadword Win32 FILETIME value, seems to be creation time of cookie
//	- a quadword Win32 FILETIME value, seems to be time of last site visit
//	- a list of dwords (length seems to be fixed to 20 dwords)
//	    dword at record offset 52 (hex 34) points to the cookie name, seems to be always 104 (hex 68)
//	    dword at record offset 60 (hex 3C) points to the file name, position depends on length of cookie name
//	    dword at record offset 84 (hex 54) seems to be a count of site visits
//	    I do not yet know the meaning of the other data in this list of dwords,
//	    I'm not even sure if they should all be interpreted as dwords
//	- the cookie name (variable length, prefixed with "Cookie:")
//	- a null byte terminating the cookie name
//	- variable number of fill bytes up to next 4-byte boundary
//	- the cookie file name (variable length)
//	- a null byte terminating the file name
//	- variable number of fill bytes up to end of record
//
// About cookie files:
//	Internet Explorer saves cookies in files containing a 9-line section for each item as follows:
//	- itemName
//	- itemValue
//	- itemDomain
//	- flags
//	- low  order value of expiry time
//	- high order value of expiry time
//	- low  order value of creation time
//	- high order value of creation time
//	- an asterisk
//	I have not yet discovered what the meaning of each flags bit could be, maybe sometime...
//	I know about 2 flag bits: 'secure' and 'httponly', but I do not yet know what their bitmask is.
//	The low and high order time values together form a Win32 FILETIME value.
//
// About Win32 FILETIME values:
//	it is the number of 100 nanosecond intervals since 01-JAN-1601 00:00:00.0000000 UTC
//	(found on http://codespeak.net/svn/wwwsearch/ClientCookie/trunk/ClientCookie/_MSIECookieJar.py)
//	Note: the JS Date Object stores milliseconds since 01-JAN-1970 00:00:00.000 UTC (the "Unix epoch")
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// ===========================================================
// We'll need a WshShell Object as well as a FileSystemObject:
// -----------------------------------------------------------
	var WSH = new ActiveXObject ("WScript.Shell")
	var FSO = new ActiveXObject ("Scripting.FileSystemObject")

// =================================================================================
// The stuff below is for correct handling of binary data stored in a String object.
// See my HRC4 module for more info. It is used for interpreting the index.dat file.
// ---------------------------------------------------------------------------------
	String.ByteChars = ""
	String.prototype.byteAt  = function(i)	{var c = String.ByteChars.indexOf(this.charAt(i)); return (c < 0 ? this.charCodeAt(i) : c+128) }
	String.prototype.wordAt  = function(i)	{return this.byteAt(i+1) * 0x100   + this.byteAt(i) }
	String.prototype.dwordAt = function(i)	{return this.wordAt(i+2) * 0x10000 + this.wordAt(i) }

// =====================================
// Some other string processing methods:
// -------------------------------------
	String.prototype.repeat = function(n)	// repeat own value n times by self-concatenation
	{	return (new Array(n+1)).join(this)
	}
	String.prototype.rpad = function(w,c)	// right pad with c to length w
	{	var r = "" + this, p = (arguments.length > 1 ? "" + c : "")
		if (p == "") p = " "
		for (; r.length < w; r += p);
		return r
	}
	String.prototype.lpad = function(w,c)	// left pad with c to length w
	{	var r = "" + this, p = (arguments.length > 1 ? "" + c : "")
		if (p == "") p = " "
		for (; r.length < w; r = p + r);
		return r
	}

// ================================================================================================
// The method below converts times as stored in Internet Explorer Cookie files to JavaScript Dates.
// And I prefer another date format: Day dd-Mon-yyyy HH:mm:ss UTC+hhmm
// ------------------------------------------------------------------------------------------------
	Date.WIN32ZEROTIME = Date.UTC(1601,0,1,0,0,0,0)
	Date.newFromCookie = function (lo, hi) {return new Date (Date.WIN32ZEROTIME + 1e-4 * (hi * 0x100000000 + lo) ) }
	Date.prototype.toString = function()
	{	var d = this.toDateString().split(" ")
		return [d[0],[("0"+d[2]).slice(-2),d[1],d[3]].join("-"),this.toTimeString()].join(" ")
	}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// ===================================
// Constructor of HRCookieItem object:
// -----------------------------------
	HRCookieItem = function (itemName, itemValue, itemDomain, itemFlags, itemExpires, itemCreated)
	{	this.init       (itemName, itemValue, itemDomain, itemFlags, itemExpires, itemCreated)
	}
// ============================================
// Static property used by the toString method:
// --------------------------------------------
	HRCookieItem.indent = " ".repeat(4)
// =============================================
// Initialisation method of HRCookieItem object:
// ---------------------------------------------
	HRCookieItem.prototype.init = function (itemName, itemValue, itemDomain, itemFlags, itemExpires, itemCreated)
	{	this.name    = itemName
		this.value   = itemValue
		this.domain  = itemDomain
		this.flags   = itemFlags
		this.expires = itemExpires
		this.created = itemCreated
	}
// =======================================
// toString method of HRCookieItem object:
// ---------------------------------------
	HRCookieItem.prototype.toString = function()
	{	var result = []
		result.push("name:    "+this.name)
		result.push("value:   "+this.value)
		result.push("domain:  "+this.domain)
		result.push("flags:   "+this.flags+" = 0x"+this.flags.toString(16) )
		result.push("expires: "+this.expires)
		result.push("created: "+this.created)
		return HRCookieItem.indent + result.join("\n"+HRCookieItem.indent)
	}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// ===============================
// Constructor of HRCookie object:
// -------------------------------
	HRCookie = function (name, path, info, parent)
	{	this.init   (name, path, info, parent)
	}
// =======================
// Some static properties:
// -----------------------
	HRCookie.indent        = " ".repeat(2)		// used by the toString method
	HRCookie.showDetails   = true			// false = ignore cookie details, i.e. index.dat info only
	HRCookie.showCsv       = false			// true implies "no details", output in csv format
	HRCookie.defaultCompareOrder = "NCLV"		// Name,Created,Lastupdate,Visitcount
	HRCookie.compareOrder  = HRCookie.defaultCompareOrder
// =============================================================================================
// in CSV output we use the list separator character as defined in the user's Regional Settings:
// ---------------------------------------------------------------------------------------------
	HRCookie.listSeparator = WSH.RegRead("HKCU\\Control Panel\\International\\sList")
// =========================================
// Initialisation method of HRCookie object:
// -----------------------------------------
	HRCookie.prototype.init = function (name, path, info, parent)
	{	var fileObject,fileStream,line,lineno,lotim,hitim
		var itemName,itemValue,itemDomain,itemFlags,itemExpires,itemCreated
		var ForReading    = 1
		this.parent       = parent
		this.name         = name
		this.time0        = Date.newFromCookie (info[0], info[1] )
		this.time1        = Date.newFromCookie (info[2], info[3] )
		this.info         = info.slice(4)
		this.filePath     = path
		this.fileExists   = FSO.FileExists (this.filePath)
		this.fileCreated  = null
		this.fileModified = null
		this.content      = []
		if (this.fileExists && HRCookie.showDetails)
		{	fileObject = FSO.GetFile (this.filePath)
			this.fileCreated  = new Date (Date.parse (fileObject.DateCreated     ) )
			this.fileModified = new Date (Date.parse (fileObject.DateLastModified) )
			fileStream  = fileObject.OpenAstextStream (ForReading)
			for (lineno = 0; !fileStream.AtEndOfStream; lineno++)
			{	line = fileStream.ReadLine()
				if (lineno%9 == 0) itemName   = line
				if (lineno%9 == 1) itemValue  = line
				if (lineno%9 == 2) itemDomain = line
				if (lineno%9 == 3) itemFlags  = parseInt(line,10)
				if (lineno%9 == 4)  lotim = parseInt(line,10)
				if (lineno%9 == 5) {hitim = parseInt(line,10); itemExpires = Date.newFromCookie(lotim,hitim) }
				if (lineno%9 == 6)  lotim = parseInt(line,10)
				if (lineno%9 == 7) {hitim = parseInt(line,10); itemCreated = Date.newFromCookie(lotim,hitim) }
				if (lineno%9 == 8) this.content.push (new HRCookieItem (itemName,itemValue,itemDomain,itemFlags,itemExpires,itemCreated) )
			}
			fileStream.Close()
		}
	}
// ===================================
// toString method of HRCookie object:
// -----------------------------------
	HRCookie.prototype.toString = function()
	{	var result = []
		if (HRCookie.showDetails)
		{	result.push("cookie name:     " + this.name)
			result.push("created:         " + this.time0)
			result.push("last site visit: " + this.time1)
			result.push("no. of visits:   " + this.info[15] )
		//	result.push("info array:      " + this.info)
			result.push("file path:       " + (this.fileExists ? "" : "NOT FOUND: ") + this.filePath)
			result.push("file created:    " + this.fileCreated)
			result.push("file modified:   " + this.fileModified)
			result.push("content:" + (this.content.length > 0 ? "\n"+this.content.join("\n\n") : "") )
			return HRCookie.indent + result.join("\n"+HRCookie.indent)
		} else if (HRCookie.showCsv)
		{	result.push(this.name)
			result.push(this.time0)
			result.push(this.time1)
			result.push(this.info[15])
			result.push(this.fileExists ? "yes" : "no")
			return result.join(HRCookie.listSeparator)
		} else
		{	result.push(this.name.rpad(this.parent.maxLength) )
			result.push(this.time0)
			result.push(this.time1)
			result.push((""+this.info[15]).lpad(5) )
			result.push(this.fileExists ? "yes" : "no")
			return result.join("  ")
		}
	}
// ====================================
// static method for comparing cookies:
// ------------------------------------
	HRCookie.compare = function (cookie1, cookie2)
	{	var i,val1,val2
		for (i = 0; i < HRCookie.compareOrder.length; i++)
		{	switch (HRCookie.compareOrder.charAt(i).toUpperCase())
			{	case "N": val1 = cookie1.name;		val2 = cookie2.name;		break;
				case "C": val1 = cookie1.time0;		val2 = cookie2.time0;		break;
				case "L": val1 = cookie1.time1;		val2 = cookie2.time1;		break;
				case "V": val1 = cookie1.info[15];	val2 = cookie2.info[15];	break;
			}
			if (val1 < val2) return -1
			if (val1 > val2) return +1
		}
		return 0
	}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// ===================================
// Constructor of HRCookieList object:
// -----------------------------------
	HRCookieList = function (filterRE)
	{	if (arguments.length < 1)
			this.init ()
		else	this.init (filterRE)
	}
// =============================================
// Initialisation method of HRCookieList object:
// ---------------------------------------------
	HRCookieList.prototype.init = function (filterRE)
	{	var i,k,m,n
		var indexObject,indexStream,indexContent,recpos,recsig,recsiz,rectxt,isCookie
		var info,cookieName,fileName,filePath,folderObject,fileCollection,fileObject
		var ForReading     = 1
		var nulbyt         = String.fromCharCode (0)
		var endrec         = String.fromCharCode (0,0,0,0)
		var registryValue  = "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders\\Cookies"
		this.timestamp     = new Date()
		this.folderPath    = WSH.ExpandEnvironmentStrings (WSH.RegRead (registryValue) )
		this.indexFileName = "index.dat"
		this.filter        = filterRE
		this.cookieList    = []
		n                  = 0
		this.maxLength     = 0
		indexObject  = FSO.GetFile (FSO.BuildPath (this.folderPath, this.indexFileName) )
		indexStream  = indexObject.OpenAsTextStream (ForReading)
		indexContent = indexStream.Read(indexObject.Size)	// read entire index.dat file at once (at cost of memory)
		indexStream.Close()
		if (indexContent.slice(0,32).search(/Client UrlCache MMF Ver \d\.\d.*/) == 0)
		{	//recpos = indexContent.dwordAt(32)
			// it seems not to work when just going to the next record using recpos and recsiz
			// so instead I now simply search for the "URL " signature and check the record content
			recpos = indexContent.indexOf("URL ")
			do {	recsig = indexContent.substr (recpos,4)
				recsiz = indexContent.dwordAt(recpos+4)*128
				if (recsig == "URL ")
				{	rectxt  = indexContent.substr(recpos,recsiz)
					k = rectxt.dwordAt(52), cookieName = rectxt.slice(k+7,rectxt.indexOf(nulbyt,k) ) // 7 = "Cookie:".length
					m = rectxt.dwordAt(60), fileName   = rectxt.slice(m  ,rectxt.indexOf(nulbyt,m) )
					isCookie = (rectxt.slice(k,k+7) == "Cookie:")
					if (isCookie && (arguments.length < 1 || cookieName.search (filterRE) >= 0) )
					{	for (i = 8, info = []; i < k; info.push(rectxt.dwordAt(i)), i += 4);
						filePath = FSO.BuildPath (this.folderPath, fileName)
						this.cookieList[n++] = new HRCookie (cookieName, filePath, info, this)
						this.maxLength = Math.max (this.maxLength, cookieName.length)
					}
				}
				//recpos += recsiz
				recpos = indexContent.indexOf("URL ",recpos+4)
			} while (recpos >= 0) //(recpos < indexContent.length && recsig != endrec)
		}
	}
// =======================================
// toString method of HRCookieList object:
// ---------------------------------------
	HRCookieList.prototype.toString = function()
	{	var result = []
		if (HRCookie.showDetails)
		{	result.push("timestamp: " + this.timestamp)
			result.push("folder:    " + this.folderPath)
			result.push("filter:    " + this.filter)
			result.push("cookies:" + (this.cookieList.length > 0 ? "\n"+this.cookieList.join("\n\n") : "") )
			result.push("" + this.cookieList.length + " cookies in " + this.indexFileName)
			return result.join("\n")
		} else
		{	var hsplit = "cookie/created/lastvisit/count/fileexists".split("/")
			if (!HRCookie.showCsv)
			{	hsplit[0] = hsplit[0].rpad(this.maxLength)
				hsplit[1] = hsplit[1].rpad(33)
				hsplit[2] = hsplit[2].rpad(33)
				header = hsplit.join("  ")
			} else	header = hsplit.join(HRCookie.listSeparator)
			return [header,this.cookieList.join("\n")].join("\n")
		}
	}
// ===================================
// sort method of HRCookieList object:
// -----------------------------------
	HRCookieList.prototype.sort = function (compareOrder)
	{	if (arguments.length > 0)
			HRCookie.compareOrder = compareOrder
		else	HRCookie.compareOrder = HRCookie.defaultCompareOrder
		this.cookieList.sort (HRCookie.compare)
	}

////////////////////////////////////////////////////////////////////////////////////////////////////////////
// end of file hrcookie.js -- (c) Henk Reints
